@@ -93,6 +93,7 @@ struct service_auth {
 	service_auth_cb cb;
 	void *user_data;
 	struct btd_device *device;
+	struct btd_adapter *adapter;
 };
 
 struct btd_adapter {
@@ -116,6 +117,7 @@ struct btd_adapter {
 	DBusMessage *discovery_cancel;	/* discovery cancel message request */
 	GSList *passkey_agents;
 	struct agent *agent;		/* For the new API */
+	guint auth_idle_id;		/* Ongoing authorization */
 	GSList *connections;		/* Connected devices */
 	GSList *devices;		/* Devices structure pointers */
 	GSList *mode_sessions;		/* Request Mode sessions */
@@ -662,8 +664,10 @@ static void session_remove(struct session_req *req)
 		g_slist_free(adapter->oor_devices);
 		adapter->oor_devices = NULL;
 
-		if (adapter->scheduler_id)
+		if (adapter->scheduler_id) {
 			g_source_remove(adapter->scheduler_id);
+			adapter->scheduler_id = 0;
+		}
 
 		adapter_ops->stop_discovery(adapter->dev_id);
 	}
@@ -1097,7 +1101,7 @@ void adapter_remove_device(DBusConnection *conn, struct btd_adapter *adapter,
 		device_set_agent(device, NULL);
 	}
 
-	device_remove(device, conn, TRUE);
+	device_remove(device, TRUE);
 }
 
 struct btd_device *adapter_get_device(DBusConnection *conn,
@@ -2388,6 +2392,9 @@ static void adapter_free(gpointer user_data)
 
 	debug("adapter_free(%p)", adapter);
 
+	if (adapter->auth_idle_id)
+		g_source_remove(adapter->auth_idle_id);
+
 	g_free(adapter->path);
 	g_free(adapter);
 }
@@ -2461,7 +2468,7 @@ void adapter_remove(struct btd_adapter *adapter)
 	debug("Removing adapter %s", adapter->path);
 
 	for (l = adapter->devices; l; l = l->next)
-		device_remove(l->data, connection, FALSE);
+		device_remove(l->data, FALSE);
 	g_slist_free(adapter->devices);
 
 	unload_drivers(adapter);
@@ -2878,6 +2885,18 @@ static void agent_auth_cb(struct agent *agent, DBusError *derr,
 	auth->cb(derr, auth->user_data);
 }
 
+static gboolean auth_idle_cb(gpointer user_data)
+{
+	struct service_auth *auth = user_data;
+	struct btd_adapter *adapter = auth->adapter;
+
+	adapter->auth_idle_id = 0;
+
+	auth->cb(NULL, auth->user_data);
+
+	return FALSE;
+}
+
 static int btd_adapter_authorize(struct btd_adapter *adapter,
 					const bdaddr_t *dst,
 					const char *uuid,
@@ -2900,24 +2919,8 @@ static int btd_adapter_authorize(struct btd_adapter *adapter,
 	if (!g_slist_find(adapter->connections, device))
 		return -ENOTCONN;
 
-	trusted = read_trust(&adapter->bdaddr, address, GLOBAL_TRUST);
-
-	if (trusted) {
-		cb(NULL, user_data);
-		return 0;
-	}
-
-	device = adapter_find_device(adapter, address);
-	if (!device)
-		return -EPERM;
-
-	agent = device_get_agent(device);
-
-	if (!agent)
-		agent = adapter->agent;
-
-	if (!agent)
-		return -EPERM;
+	if (adapter->auth_idle_id)
+		return -EBUSY;
 
 	auth = g_try_new0(struct service_auth, 1);
 	if (!auth)
@@ -2926,12 +2929,33 @@ static int btd_adapter_authorize(struct btd_adapter *adapter,
 	auth->cb = cb;
 	auth->user_data = user_data;
 	auth->device = device;
+	auth->adapter = adapter;
+
+	trusted = read_trust(&adapter->bdaddr, address, GLOBAL_TRUST);
+
+	if (trusted) {
+		adapter->auth_idle_id = g_idle_add_full(G_PRIORITY_DEFAULT_IDLE,
+							auth_idle_cb, auth,
+							g_free);
+		return 0;
+	}
+
+	agent = device_get_agent(device);
+
+	if (!agent)
+		agent = adapter->agent;
+
+	if (!agent) {
+		g_free(auth);
+		return -EPERM;
+	}
 
 	dev_path = device_get_path(device);
 
 	err = agent_authorize(agent, dev_path, uuid, agent_auth_cb, auth, g_free);
-
-	if (err == 0)
+	if (err < 0)
+		g_free(auth);
+	else
 		device_set_authorizing(device, TRUE);
 
 	return err;
@@ -2987,6 +3011,12 @@ int btd_cancel_authorization(const bdaddr_t *src, const bdaddr_t *dst)
 	if (!device)
 		return -EPERM;
 
+	if (adapter->auth_idle_id) {
+		g_source_remove(adapter->auth_idle_id);
+		adapter->auth_idle_id = 0;
+		return 0;
+	}
+
 	/*
 	 * FIXME: Cancel fails if authorization is requested to adapter's
 	 * agent and in the meanwhile CreatePairedDevice is called.
